/*
 * (C) Copyright Keith Visco 1999  All rights reserved.
 *
 * The contents of this file are released under an Open Source 
 * Definition (OSD) compliant license; you may not use this file 
 * execpt in compliance with the license. Please see license.txt, 
 * distributed with this file. You may also obtain a copy of the
 * license at http://www.clc-marketing.com/xslp/license.txt
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * The Copyright owner will not be liable for any damages suffered by
 * you as a result of using the Program. In no event will the Copyright
 * owner be liable for any special, indirect or consequential damages or
 * lost profits even if the Copyright owner has been advised of the
 * possibility of their occurrence.
 *
 * $Id$
 */
package com.kvisco.xpath;

import com.kvisco.util.QuickStack;
import com.kvisco.util.Iterator;

import java.util.Hashtable;


import org.w3c.dom.*;

/**
 * The default implementation for ExprContext
 * @author <a href="kvisco@ziplink.net">Keith Visco</a>
 * @version $Revision$ $Date$
**/
public class ExprContextImpl implements ExprContext {

    
      //----------------------/
     //- Instance Variables -/
    //----------------------/

    /**
     * A hastable for referencing the parents of
     * nodes, that do not have parents
    **/
    private Hashtable  parents            = null;
    /**
     * A pointer to the next node to check, 
     * when indexing attributes
    **/
    private Hashtable  indexingInfo            = null;
    
    /**
     * keeps track of the document order for a node
    **/
    private Hashtable documentOrders     = null;
    
    /**
     * ID Reference table for Source tree
    **/
    private Hashtable idReferences      = null;
    
    private QuickStack nodeSetStack     = null;
    
    
    private VariableSet _variableSet    = null;
    
      //----------------/
     //- Constructors -/
    //----------------/
    
    /**
     * Creates a new ExprContextImpl
    **/
    public ExprContextImpl() {
        super();
        parents           = new Hashtable();
        documentOrders    = new Hashtable();
        indexingInfo      = new Hashtable();
        idReferences      = new Hashtable();
        nodeSetStack      = new QuickStack();
        _variableSet      = new VariableSet(1);
    } //-- ExprContextImpl

      //------------------/
     //- Public Methods -/
    //------------------/
    
    public void addIdReference(String id, Node node) {
        idReferences.put(id, node);
    } //-- addIdReference
    
    /**
     * Finds the parent of a given Node. Since certain
     * Nodes in DOM level 1 do not have parents (theoretically)
     * we need a way to find the parent of any node (mainly
     * attribute nodes).
     * @param node the Node to find the parent of
     * @return the parent of the given node
     * @see com.kvisco.xpath.ExprContext
    **/
    public Node findParent(Node node) {
        return getParentNode(node);
    } //-- findParent
    
    /**
     * Determines the document order of a given node.
     * Document Order is defined by
     * the document order of the parent of the given node and 
     * the childNumber of the given Node.
     * The document order for a Document node is 0.
     * @see com.kvisco.xpath.ExprContext
    **/
    public int[] getDocumentOrder(Node node) {
        
        int[] order = null;
        
        if (node == null) {
            order = new int[1];
            order[0] = -1;
            return order;
        }
        
        //-- check cache
        //-- * due to bugs in XML4J 1.1.x (2.x works fine) 
        //-- * we need to use the System.identityHash to
        //-- * create a unique key. The problem is Attr nodes
        //-- * with the same name, generate the same hash code.
        Object key = createKey(node);
        
        order = (int[]) documentOrders.get(key);
        if (order != null) return order;
        
        
        Node parent = null;
        //-- calculate document order
        if (node.getNodeType() == Node.ATTRIBUTE_NODE) {
            // Use parent's document order for attributes
            parent = findParent((Attr)node);
            if (parent == null) {
                order = new int[3];
                order[0] = 0;
                order[1] = 0;
                order[2] = childNumber(node);
            }
            else { 
                int[] porder = getDocumentOrder(parent);
                order = new int[porder.length+2];
                System.arraycopy(porder, 0, order,0, porder.length);
                order[order.length-2] = 0;
                order[order.length-1] = childNumber(node);
            }
        }
        else if (node.getNodeType() == Node.DOCUMENT_NODE) {
            order = new int[1];
            order[0] = 0;
        }
        else {
            
            //-- get parent's document order
            parent = getParentNode(node);
            int[] porder = getDocumentOrder(getParentNode(node));
            order = new int[porder.length+1];
            System.arraycopy(porder, 0, order,0, porder.length);
            order[order.length-1] = childNumber(node);
        }
        
        //-- add to cache
        documentOrders.put(key,order);
        return order;
    } //-- getDocumentOrder
    
    
    /**
     * Returns the current context NodeSet
     * @return the current context NodeSet
     * @see com.kvisco.xpath.ExprContext
    **/
    public NodeSet getNodeSet() {
        if (nodeSetStack.empty()) return null;
        return (NodeSet)nodeSetStack.peek();
    } //-- getNodeSet

    /**
     * Adds the given NodeSet set to the top of the NodeSet stack
     * @param nodeSet the NodeSet to put on the top of the Context NodeSet
     * stack
     * @see com.kvisco.xpath.ExprContext
    **/
    public void pushNodeSet(NodeSet nodeSet) {
        nodeSetStack.push(nodeSet);
    } //-- pushNodeSet
    
    /**
     * Removes the top NodeSet from the NodeSet stack
     * @return the NodeSet that was "popped" off the top of the NodeSet stack
     * @see com.kvisco.xpath.ExprContext
    ***/
    public NodeSet popNodeSet() {
        return (NodeSet)nodeSetStack.pop();
    } //-- popNodeSet
    
    /**
     * Returns a call to the Function that has the given name. This method
     * is used for the XPath Extension Functions.
     * @return the FunctionCall for the Function with the given name
     * @see com.kvisco.xpath.ExprContext
    **/
    public FunctionCall resolveFunctionCall(String functionName) {
        return null;
    } //-- resolveFunctionCall
    
    /**
     * Returns the Node that is associated with the given Id.
     * @return the Node that is associated with the given Id, or null
     * if no Node has been found
     * @see com.kvisco.xpath.ExprContext
    **/
    public Node resolveId(String id) {
        return (Node)idReferences.get(id);
    } //-- resolveId
    
    
    /**
     * Retrieves the defined stylesheet Variable with the given name
     * @param name the name of the Variable to get
     * @return the Variablet with the specified name, or null if not found
     * @see com.kvisco.xpath.ExprContext
    **/
    public ExprResult getVariable(String name) {
        if (_variableSet != null) {
            return _variableSet.getValue(name);
        }
        return null;
    } //-- getVariable
    
    
    public VariableSet getVariableSet() {
        return this._variableSet;
    } //-- getVariabelSet
    
    
    /**
     * Sets the Id references for this ExprContext
     * @param idRefs the set of Id references used for Resolving #IDREF
    **/
    public void setIdReferences(Hashtable idRefs) {
        this.idReferences = idRefs;
    } //-- setIdReferences
    
    public void setVariableSet(VariableSet variableSet) {
        this._variableSet = variableSet;
    } //-- setVariableSet(VariableSet)
    
     
    //-------------------/
    //- Private Methods -/
    //-------------------/
    
    /**
     * Returns the child number of a Node
     * @param node the Node to retrieve the child number of
     * @return the child number of the given node
    **/
    private int childNumber(Node node) {
        int c = 0;
        Node tmpNode = node;
        while ((tmpNode = tmpNode.getPreviousSibling())!=  null) ++c;
        return c;
    } //-- childNumber
    
    /**
     * This method is only here because XML4J 1.1.x has a nasty
     * bug with it's Attr Node implementation. All Attr nodes with
     * the same name, hash to the same value. This has been fixed
     * in their 2.x version.
    **/
    private Object createKey(Node node) {
        return new Integer(System.identityHashCode(node));
    } //-- createKey
    
    
    /**
     *
    **/
    private Node getParentNode(Node node) {
        
        if (node == null) return null;
        switch (node.getNodeType()) {
            case Node.ATTRIBUTE_NODE:
                break;
            default:
                return node.getParentNode();
        } 
        Attr attr = (Attr)node;
        
        if (attr == null) return null;
        
        //-- due to bugs in XML4J 1.1.x we need to
        //-- create a unique key for attr
        Node parent = (Node)parents.get(createKey(attr));
        
        if (parent != null) return parent;
        
        //-- getIndexingInfo
        Document doc = attr.getOwnerDocument();
        IndexState idxState = (IndexState)indexingInfo.get(doc);
        if (idxState == null) {
            indexingInfo.put(doc, (idxState = new IndexState()));
            idxState.next = doc.getDocumentElement();
            if (idxState.next == null) idxState.done = true;
        }
        if (idxState.done) return null; //-- yikes no parent found!
        
        
        boolean found = false;
        boolean alreadyProcessed  = false;
        while (!found) {
            
            if (idxState.next == null) {
                idxState.done = true;
                break;
            }
            if (!alreadyProcessed) {
                //-- index attributes
                if (idxState.next.getNodeType() == Node.ELEMENT_NODE) {
                    Element element = (Element)idxState.next;
                    NamedNodeMap atts = element.getAttributes();
                    if (atts != null) {
                        for (int i = 0; i < atts.getLength(); i++) {
                            Node tmpNode = atts.item(i);
                            //-- due to bugs in XML4J 1.1.x we need to
                            //-- create a unique key for attr
                            parents.put(createKey(tmpNode), element);
                            if (attr == tmpNode) {
                                found = true;
                                parent = element;
                            }
                        }
                    }
                }
            }
            
            //-- set next node to check
            if ((!alreadyProcessed) && (idxState.next.hasChildNodes())) {
                Node child = idxState.next.getFirstChild();
                idxState.next = child;
            }
            else if (idxState.next.getNextSibling() != null) {
                idxState.next = idxState.next.getNextSibling();
                alreadyProcessed = false;
            }
            else {
                idxState.next = getParentNode(idxState.next);
                //-- already checked parent, now check sibling
                alreadyProcessed = true;
            }
        }
        return parent;
    } //-- findParent;
    
    class IndexState {
        Node next = null;
        boolean done = false;
        protected IndexState() {
            super();
        }
    }
        
} //-- ExprContextImpl