/*
 * (C) Copyright Keith Visco 1999  All rights reserved.
 *
 * The contents of this file are released under an Open Source 
 * Definition (OSD) compliant license; you may not use this file 
 * execpt in compliance with the license. Please see license.txt, 
 * distributed with this file. You may also obtain a copy of the
 * license at http://www.clc-marketing.com/xslp/license.txt
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * The Copyright owner will not be liable for any damages suffered by
 * you as a result of using the Program. In no event will the Copyright
 * owner be liable for any special, indirect or consequential damages or
 * lost profits even if the Copyright owner has been advised of the
 * possibility of their occurrence.
 *
 * $Id$
 */

package com.kvisco.xpath;

/**
 * The set of parameters that can be passed to an XPath Function
 * @author <a href="mailto:kvisco@ziplink.net">Keith Visco</a>
 * @version $Revision$ $Date$
**/
public class ParamList {
    
    private int DEFAULT_SIZE = 3;
    
    private Expr[] params;
    
    private int initialSize = DEFAULT_SIZE;
    
    /**
     * The next available location in the expr array
    **/
    private int paramCount = 0;
    
    /**
     * Creates a new parameter list with the default Size
    **/
    public ParamList() {
        params = new Expr[DEFAULT_SIZE];
    } //-- ParamList
    
    /**
     * Creates a new parameter list with the specified size
     * @param size the number of parameters this list will hold
    **/
    public ParamList(int size) {
        initialSize = size;
        params = new Expr[size];
    } //-- ParamList
    
    /**
     * Adds the specified Expr to the list
     * @param expr the Expr to add to the list
     * @return true if the Expr is added to the list
    **/
    public boolean addParameter(Expr expr) {
        if (paramCount == params.length) increaseSize();
        params[paramCount++] = expr;
        return true;
    } //-- add

    /**
     * Adds the specified Expr to the list at the specified index
     * @param expr the Expr to add to the list
     * @param index the location in the param list to add the expr 
     * @return true if the Expr is added to the list
     * @exception IndexOutOfBoundsException
    **/
    public boolean addParameter(int index, Expr expr) 
        throws IndexOutOfBoundsException 
    {
        if ((index < 0) || (index > paramCount)) 
            throw new IndexOutOfBoundsException();
            
        // make sure we have room to add the object
        if (paramCount == params.length) increaseSize();
        
        if (index == paramCount) {
            params[paramCount++] = expr;
        }
        else {
            shiftUp(index);
            params[index] = expr;
            ++paramCount;
        }
        return true;
    } //-- add
    
    /**
     * Removes all expressions from the list
    **/
    public void clear() {
        for (int i = 0; i < paramCount; i++) {
            params[i] = null;
        }
        paramCount = 0;
    } //-- clear
    
    /**
     * Returns the Expr at the specified position in this list.
     * @param index the position of the Expr to return
     * @exception IndexOutOfBoundsException 
    **/
    public Expr getParameter(int index) throws IndexOutOfBoundsException {
        if ((index < 0) || index >= paramCount) {
            throw new IndexOutOfBoundsException();
        }
        return params[index];
    } //-- get
    
    /**
     * Returns true if there are no expressions in the List.
     * @return true if there are no expressions in the List.
    **/
    public boolean isEmpty() {
        return (paramCount == 0);
    } //-- isEmpty

    /**
     * Removes the Expr at the specified index from the List
     * @param index the position in the list to remove the Expr from
     * @return the Expr that was removed from the list
    **/
    public Expr removeParameter(int index) {
        
        if ((index < 0) || (index > paramCount)) return null;
        Expr expr = params[index];
        shiftDown(index+1);
        --paramCount;
        return expr;
    } //-- remove

    /**
     * Removes the first occurrence of the specified Expr from the List
     * @param expr the Expr to remove from the List
     * @return true if the Expr was removed from the list
    **/
    public boolean removeParameter(Expr expr) {
        
        int index = -1;
        for (int i = 0; i < paramCount; i++) {
            if (params[i] == expr) {
                index = i;
                break;
            }
        }
        if (index > -1) removeParameter(index);
        else return false;
        return true;
    } //-- remove
    
    /**
     * Reduces the capacity of the internal buffer to the current size
     * freeing up unused memory.
    **/
    public void trimToSize() {
        if (params.length == paramCount) return;
        Expr[] pointer = params;
        params = new Expr[paramCount];
        System.arraycopy(pointer, 0, params, 0, paramCount);
        pointer = null;
    } //-- trimBuffer
    
    /**
     * Replaces the Expr at the specified position in this list 
     * with the specified Expr.
     * @param index the position in the list to place the Expr at
     * @param expr the Expr to add to the list
     * @return the Expr that was replaced or null
     * @exception IndexOutOfBoundsException
    **/

    public Expr setParameter(int index, Expr expr) 
        throws IndexOutOfBoundsException 
    {
        if ((index < 0) || (index > paramCount)) 
            throw new IndexOutOfBoundsException();
        Expr oldExpr = null;
        if (index == paramCount) addParameter(expr);
        else {
            oldExpr = params[index];
            params[index] = expr;
        }
        return oldExpr;
    } //-- set

     
    /**
     * Returns the number of expressions in the List
     * @return the number of expressions in the List
    **/
    public int getParameterCount() {
        return paramCount;
    } //-- size

      //-------------------/
     //- Private Methods -/
    //-------------------/
    
    /**
     * Basically the same as a vector,
     * increase the list by a factor of its initial size
    **/
    private void increaseSize() {
        Expr[] pointer = params;
        int length = (pointer.length > 0) ? pointer.length : 1;
        params = new Expr[length*2];
        System.arraycopy(pointer, 0, params, 0, pointer.length);
        pointer = null;
    } //-- increaseSize
    
    /**
     * Shifts all elements at the specified index to down by 1
    **/
    private void shiftDown(int index) {
        if ((index <= 0) || (index >= paramCount)) return;
        System.arraycopy(params, index, params, index - 1, paramCount - index);        
        // clean up for gc
        params[paramCount-1] = null;
    } //-- shiftDown
    
    /**
     * Shifts all elements at the specified index up by 1
    **/
    private void shiftUp(int index) {
        if (index == paramCount) return;
        if (paramCount == params.length) increaseSize();
        System.arraycopy(params, index, params, index + 1, paramCount - index);
    } //-- shiftUp
    
    
    
} //-- List