/*
 * (C) Copyright Keith Visco 1999  All rights reserved.
 *
 * The contents of this file are released under an Open Source 
 * Definition (OSD) compliant license; you may not use this file 
 * execpt in compliance with the license. Please see license.txt, 
 * distributed with this file. You may also obtain a copy of the
 * license at http://www.clc-marketing.com/xslp/license.txt
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * The Copyright owner will not be liable for any damages suffered by
 * you as a result of using the Program. In no event will the Copyright
 * owner be liable for any special, indirect or consequential damages or
 * lost profits even if the Copyright owner has been advised of the
 * possibility of their occurrence.
 *
 * $Id$
 */

package com.kvisco.xpath;

import org.w3c.dom.*;

/**
 * This class represents a PathExpr and LocationPath 
 * pattern
 * @author <a href="kvisco@ziplink.net">Keith Visco</a>
 * @version $Revision$ $Date$
**/
public class PathExpr implements Expr, MatchExpr {

    public final int ABSOLUTE = 0;
    public final int RELATIVE = 1;
    
    
    /**
     * The local FilterExpr of this PathExpr
    **/
    private FilterBase _filter = null;
    
    /**
     * The subpath of this PathExpr
    **/
    private PathExpr _subPath = null;
    
    /**
     * Creates a new PathExpr with will not match any node, and will evaluate
     * to an empty NodeSet
    **/
    public PathExpr() {
        super();
    } //-- PathExpr
    
    /**
     * Creates a new PathExpr with the given FilterBase
    **/
    public PathExpr(FilterBase filterBase) {
        super();
        _filter = filterBase;
    } //-- PathExpr

    /**
     * Creates a new PathExpr with the given FilterBase and PathExpr
    **/
    public PathExpr(FilterBase filterBase, PathExpr pathExpr) {
        super();
        _filter = filterBase;
        _subPath = pathExpr;
    } //-- PathExpr
    
    
      //------------------/    
     //- Public Methods -/
    //------------------/    
    
    /**
     * Returns the type of Expr this Expr represents
     * @return the type of Expr this Expr represents
    **/
    public short getExprType() {
        return Expr.NODE_SET;
    } //-- getExprType
    
    /**
     * Evaluates this Expr using the given context Node and ExprContext
     * @param context the current context Node
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * environment
     * @return the ExprResult
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    public ExprResult evaluate(Node context, ExprContext exprContext)
        throws InvalidExprException 
    {
        return evaluate(context, exprContext, true);
    } //-- evaluate
    
    /**
     * Evaluates this Expr using the given context Node and ExprContext
     * @param context the current context Node
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * environment
     * @return the ExprResult
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    private NodeSet evaluate
        (Node context, ExprContext exprContext, boolean start)
        throws InvalidExprException 
    {
        
        
        if ((_filter == null) || (context == null))
            return new NodeSet(0);

        if (start) {
            if ((isAbsolute()) && (context.getNodeType() != Node.DOCUMENT_NODE))
                context = context.getOwnerDocument();
        }
        
        boolean ancestorMode = (_filter.getAncestryOp() == FilterBase.ANCESTOR_OP);
            
        NodeSet nodes = _filter.evaluate(context, exprContext);
        if (ancestorMode)
            nodes.add(fromDescendants(_filter, context, exprContext));
            
        if (_subPath != null) {
            NodeSet tmpNodes = new NodeSet();
            exprContext.pushNodeSet(nodes);
            for (int j = 0; j < nodes.size(); j++) {
                Node node = nodes.get(j);
                tmpNodes.add(_subPath.evaluate(node, exprContext, false));
            }
            exprContext.popNodeSet();
            nodes = tmpNodes;
        }
        return nodes;
    } //-- evaluate(context, ExprContext, boolean)
    
    private static NodeSet fromDescendants
        (FilterBase filterBase, Node context, ExprContext exprContext) 
        throws InvalidExprException
    {
        NodeSet nodeSet = new NodeSet();
        
        if (context == null) return nodeSet;
        
        NodeList nl = context.getChildNodes();
        
        for (int i = 0; i < nl.getLength(); i++) {
            Node child = nl.item(i);
            
            //-- filter DOCTYPE and entity elements
            switch (child.getNodeType()) {
                case Node.ENTITY_NODE:
                case Node.DOCUMENT_TYPE_NODE:
                    continue;
                default:
                    break;
            }
            nodeSet.add(filterBase.evaluate(child, exprContext));
            //-- check childs descendants
            if (child.hasChildNodes()) {
                NodeSet temp = fromDescendants(filterBase, child, exprContext);
                nodeSet.add(temp);
            }
        }
        return nodeSet;
        
    } //-- fromDecendants
    
    public FilterBase getFilter() {
        return _filter;
    }
    public void setFilter(FilterBase filterBase) {
        this._filter = filterBase;
    } //-- setFilter
    
    public void setSubPath(PathExpr pathExpr) {
        this._subPath = pathExpr;
    } //-- setSubPath
    
    /**
     * Returns the String representation of this PathExpr
     * @return the String representation of this PathExpr
    **/
    public String toString() {
        return toString(new StringBuffer());
    } //-- toString

    private String toString(StringBuffer sb) {
        if (sb == null) sb = new StringBuffer();
        if (_filter != null) {
            switch (_filter.getAncestryOp()) {
                case FilterBase.ANCESTOR_OP:
                    sb.append("//");
                    break;
                case FilterBase.PARENT_OP:
                    sb.append("/");
                    break;
                default:
                    break;                
            }
            sb.append(_filter.toString());
        }
        if (_subPath != null) _subPath.toString(sb);
        return sb.toString();
    } //-- toString
    
    
    
    /**
     * Determines the priority of a PatternExpr as follows:
     * <PRE>
     *  - From the 19990421 XSL Working Draft -
     *  + If the Pattern has the form of a QName optionally preceded by
     *    the @ character, then the priority is 0.
     *  + Otherwise if the pattern consists of just a NodeTest then the
     *    priority is -1
     *  + Otherwise the priority is 1
     * </PRE>
     * @return the priority for this PatternExpr
    **/
    public int getDefaultPriority() {
        if (_subPath != null) return 1;
        return _filter.getDefaultPriority();
    } //-- getDefaultPriority
    
    
    /**
     * Determines if this PathExpr Represents an AbsolutePathExpr or not
    **/
    public boolean isAbsolute() {
        if (_filter == null) return true;
        return (_filter.getAncestryOp() != FilterBase.NO_OP);
    } //-- isAbsolute
    
    /**
     * Determines if the given node is matched by this MatchExpr with
     * respect to the given context node.
     * @param node the node to determine a match for
     * @param context the Node which represents the current context
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * @return true if the given node is matched by this MatchExpr
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    public boolean matches(Node node, Node context, ExprContext exprContext)
        throws InvalidExprException 
    {
        if (node == null) return false;
        
        NodeSet nodes = reverseEvaluate(node, exprContext);
        return (nodes.size() > 0);
        
    } //-- matches
    
    private static Node getParentNode(Node node, ExprContext exprContext) {
        return exprContext.findParent(node);
    } //-- getParentNode,
    
    
    protected NodeSet reverseEvaluate(Node context, ExprContext exprContext) 
        throws InvalidExprException
    {
        
        if (_filter == null) return new NodeSet(0);
        
        NodeSet cnodes = null;
        if (_subPath != null) {
            cnodes = _subPath.reverseEvaluate(context, exprContext);
        } 
        else {
            cnodes = new NodeSet(1);
            cnodes.add(context);
        }
        
        //-- a tree depth of 7 is pretty big, but this
        //-- value is much smaller than the default size
        //-- for a nodeSet, so that's why I am setting it 
        //-- to 7, perhaps I should calculate the depth?
        NodeSet nodeSet = new NodeSet(7);
        
        for (int i = 0; i < cnodes.size(); i++) {
            Node node = cnodes.get(i);
            if (_filter.matches(node, context, exprContext) ) {
                int ancestryOp = _filter.getAncestryOp();
                switch(ancestryOp) {
                    case FilterBase.ANCESTOR_OP:
                        {
                            Node parent = node;
                            while ((parent = parent.getParentNode()) != null)
                                nodeSet.add(parent);
                        }
                        break;
                    case FilterBase.PARENT_OP:
                        {
                            Node parent = node.getParentNode();
                            if (parent != null) nodeSet.add(parent);
                        }
                        break;
                    default:
                        nodeSet.add(node);
                        break;
                }
            }
        }
        return nodeSet;
    } //--  reverseEvaluate
    
} // -- PathExpr