/*
 * (C) Copyright Keith Visco 1999  All rights reserved.
 *
 * The contents of this file are released under an Open Source 
 * Definition (OSD) compliant license; you may not use this file 
 * execpt in compliance with the license. Please see license.txt, 
 * distributed with this file. You may also obtain a copy of the
 * license at http://www.clc-marketing.com/xslp/license.txt
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * The Copyright owner will not be liable for any damages suffered by
 * you as a result of using the Program. In no event will the Copyright
 * owner be liable for any special, indirect or consequential damages or
 * lost profits even if the Copyright owner has been advised of the
 * possibility of their occurrence.
 *
 * $Id$
 */

package com.kvisco.xpath;

import org.w3c.dom.*;

//import com.kvisco.util.QuickStack;

import java.util.Hashtable;
import java.util.StringTokenizer;

/**
 * This class represents a SelectExpr
 * <PRE>
 * [1] SelectExpr ::= UnionExpr
 * </PRE>
 * @author <a href="mailto:kvisco@ziplink.net">Keith Visco</a>
 * @version $Revision$ $Date$
**/
public class SelectExpr implements Expr {
    
    /**
     * The UnionExpr for this SelectExpr
    **/
    private UnionExpr unionExpr = null;
    
      //----------------/
     //- Constructors -/
    //----------------/

    /**
     * Creates a new SelectExpr. This SelectExpr will select the 
     * empty node set by default.
    **/
    public SelectExpr() {
        super();
    } //-- SelectExpr


    public String toString() {
        if (this.unionExpr != null) return unionExpr.toString();
        else return "";
    } //-- toString
    
    
    public short getExprType() {
        return Expr.NODE_SET;
    } //-- getExprType
    
    /**
     * Evaluates this Expr using the given context Node and ExprContext
     * @param context the current context Node
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * @return the ExprResult
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    public ExprResult evaluate(Node context, ExprContext exprContext)
        throws InvalidExprException
    {
        return selectNodes(context, exprContext);
    } //-- evaluate
    
      //---------------------/
     //- Protected Methods -/
    //---------------------/
    
    protected void getUnionExpr(UnionExpr unionExpr) {
        this.unionExpr = unionExpr;
    } //-- setUnionExpr
    
    /**
     * Selects all nodes matching this SelectExpr
     * @param context the current context Node
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * @return the ExprResult
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    protected NodeSet selectNodes(Node context, ExprContext exprContext) 
        throws InvalidExprException
    {
        if (unionExpr == null) return new NodeSet(0);
        NodeSet nodes = (NodeSet)unionExpr.evaluate(context, exprContext);
        nodes = reorder(nodes, exprContext);
        return nodes;
    } // -- selectNodes
    
    
    protected void setUnionExpr(UnionExpr unionExpr) {
        this.unionExpr = unionExpr;
    } //-- setUnionExpr

      //-------------------/
     //- Private Methods -/
    //-------------------/

    
    /**
     * Compares the document orders.
    **/
    private int compareOrders(int[] orderA, int[] orderB) {
        int idx = 0;
        while ((idx < orderA.length) && (idx < orderB.length)) {
            if (orderA[idx] < orderB[idx]) return -1;
            else if (orderB[idx] < orderA[idx]) return 1;
            ++idx;
        }
        if (orderA.length < orderB.length) return -1;
        else if (orderB.length < orderA.length) return 1;
        else return 0;
    } //-- compareOrders
    
    /**
     * Reorders the array of Nodes so that the nodes are in Document Order.
    **/
    private NodeSet reorder(NodeSet nodes, ExprContext exprContext) {
        if (nodes.size() < 2) return nodes;
        
        NodeSet sorted = new NodeSet(nodes.size());
        sorted.add(nodes.get(0));
        
        for (int i = 1; i < nodes.size(); i++) {
            Node node = nodes.get(i);
            
            int[] order0 = exprContext.getDocumentOrder(node);
            for (int j = 0; j < sorted.size(); j++) {
                int[] order1 = exprContext.getDocumentOrder(sorted.get(j));
                if (compareOrders(order0, order1) < 0) {
                    sorted.add(j, node);
                    break;
                }
                else if (j == sorted.size()-1) {
                    sorted.add(node);
                    break;
                }
            }
        }
        return sorted;
    } //-- reorder
        
} //-- SelectExpr
