/*
 * (C) Copyright Keith Visco 1999  All rights reserved.
 *
 * The contents of this file are released under an Open Source 
 * Definition (OSD) compliant license; you may not use this file 
 * execpt in compliance with the license. Please see license.txt, 
 * distributed with this file. You may also obtain a copy of the
 * license at http://www.clc-marketing.com/xslp/license.txt
 *
 * The program is provided "as is" without any warranty express or
 * implied, including the warranty of non-infringement and the implied
 * warranties of merchantibility and fitness for a particular purpose.
 * The Copyright owner will not be liable for any damages suffered by
 * you as a result of using the Program. In no event will the Copyright
 * owner be liable for any special, indirect or consequential damages or
 * lost profits even if the Copyright owner has been advised of the
 * possibility of their occurrence.
 *
 */

package com.kvisco.xpath.functions;

import com.kvisco.xpath.*;
import org.w3c.dom.Node;

/**
 * A class that represents the XSL local-part(NodeSetExpr) function call
 * @author <a href="mailto:kvisco@ziplink.net">Keith Visco</a>
**/
public class XMLNamesFunctionCall extends FunctionCall {
    
    public static final short LOCAL_PART   = 1;
    public static final short NAME         = 2;
    public static final short NAMESPACE    = 3;
    
    private short fType = NAME;
    
    /**
     * Creates a new XMLNames Function Call
    **/
    public XMLNamesFunctionCall() {
        super("name");
    } //-- XMLNamesFunctionCall
    
    /**
     * Creates a new XMLNames Function Call
    **/
    public XMLNamesFunctionCall(short type) {
        super(getFunctionName(type));
        this.fType = type;
    } //-- XMLNamesFunctionCall
    
    /**
     * Evaluates this Expr using the given context Node and ExprContext
     * @param context the current context Node
     * @param exprContext the ExprContext which has additional
     * information sometimes needed for evaluating XPath expressions
     * environment
     * @return the ExprResult
     * @exception InvalidExprException when an invalid expression is
     * encountered during evaluation
    **/
    public ExprResult evaluate(Node context, ExprContext exprContext)
        throws InvalidExprException
    {
        
        NodeSet nodeSet = exprContext.getNodeSet();
        if (nodeSet == null) {
            nodeSet = new NodeSet();
            nodeSet.add(context);
        }

        String nodeName = null;
        if (getParameterCount() == 1) {
            Expr expr = getParameter(0);
            ExprResult result = expr.evaluate(context, exprContext);
            if (result.getResultType() == ExprResult.NODE_SET) {
                NodeSet nodes = (NodeSet)result;
                if (nodes.size() > 0) 
                    nodeName = nodes.get(0).getNodeName();
            }
        }
        else nodeName = context.getNodeName();
        
        switch(fType) {
            case LOCAL_PART:
                return new StringResult(getLocalPart(nodeName));
            case NAMESPACE:
                return new StringResult(getNameSpace(nodeName));
            default:
                return new StringResult(nodeName);
        }
    } //-- evaluate

    /**
     * Returns the local part of the qualified XML name
     * @param qName the qualified XML name
     * @return the local part of the qualified XML name
    **/
    private static String getLocalPart(String qName) {
        if (qName == null) return "";
        
        int idx = qName.indexOf(':');
        
        if (idx >= 0) {
            return qName.substring(idx+1);
        }
        return qName;
    } //-- getLocalPart
    
    /**
     * Returns the namespace part of the qualified XML name
     * @param qName the qualified XML name
     * @return the namespace part of the qualified XML name
    **/
    private static String getNameSpace(String qName) {
        
        if (qName == null) return "";
        
        int idx = qName.indexOf(':');
        
        if (idx > 0) {
            return qName.substring(0,idx);
        }
        return "";
    } //-- getNameSpace
    
    /**
     * 
    **/
    private static String getFunctionName(short type) {
        switch(type) {
            case LOCAL_PART:
                return "local-part";
            case NAMESPACE:
                return "namespace";
            default:
                return "name";
        }
    } //-- getFunctionName
    
} //-- XMLNamesFunctionCall